from collections import defaultdict
from typing import Union


def ddict2dict(d: Union[defaultdict, dict]) -> dict:
    """convert defaultdict into a dict"""
    for k, v in d.items():
        if isinstance(v, dict):
            d[k] = ddict2dict(v)
    return dict(d)



import functools
import logging

from contextlib import contextmanager

import attr

from dbnd._core.settings import TrackingConfig
from dbnd._core.settings.tracking_config import ValueTrackingLevel
from dbnd._core.tracking.metrics import log_dataset_op
from dbnd_airflow.tracking.airflow_connections import get_conn_path
from dbnd_airflow.tracking.config import AirflowTrackingConfig
from dbnd_snowflake.extract_sql_query import extract_from_sql
from dbnd_snowflake.snowflake_table import log_snowflake_table_targets


logger = logging.getLogger(__name__)


@contextmanager
def log_sql_target_wrapper(conn_str, conn_path, sql_query):
    # type: (str, str, str) -> None
    """
    Context manager to handle the tracking of sql query execution.

    @param conn_str:
    @param conn_path: connection string format as path. use to format a the target's path
    @param sql_query: a sql query to extract the table operations from
    """
    try:
        yield
    except Exception:
        log_sql_targets(conn_str, conn_path, sql_query, False)
        raise
    else:
        log_sql_targets(conn_str, conn_path, sql_query, True)


def log_sql_targets(conn_str, path, sql_query, is_succeed):
    try:
        for target_op in extract_from_sql(path, sql_query):
            target_op = attr.evolve(target_op, success=is_succeed)

            if (
                target_op.path.startswith("snowflake")
                and TrackingConfig.current().value_reporting_strategy
                == ValueTrackingLevel.ALL
                and not target_op.name.startswith("@")
            ):
                # 1) snowflake tables are lazy evaluated types
                # we support they been logs only if the the strategy is All
                # 2) staging is not supported yet
                log_snowflake_table_targets(
                    table_op=target_op, connection_string=conn_str
                )

            else:
                # the target can actually be a non table like S3 file that used
                # as part of the sql query
                log_dataset_op(
                    op_path=target_op.path,
                    op_type=target_op.operation,
                    success=target_op.success,
                )
    except Exception:
        logger.exception("Caught exception will trying to log target for sql_query")


def patch_airflow_db_hook(operator_cls, wrapper):
    def wrap_sql_executor(name):
        original_run = getattr(operator_cls, name)
        setattr(operator_cls, name, wrapper(original_run))

    from airflow.hooks.dbapi_hook import DbApiHook

    if issubclass(operator_cls, DbApiHook):
        wrap_sql_executor("run")  # :run(self, sql, autocommit=False, parameters=None)
        wrap_sql_executor("get_records")  # :get_records(self, sql, parameters=None)
        wrap_sql_executor("get_pandas_df")  # :get_pandas_df(self, sql, parameters=None)
        wrap_sql_executor("get_first")  # :get_first(self, sql, parameters=None)


def config_base_target_reporter(original_run):
    @functools.wraps(original_run)
    def _run_sql(self, sql, *args, **kwargs):
        config = AirflowTrackingConfig.current()

        if config.sql_reporting:
            # each instance of DbApiHook has `conn_name_attr` attribute that map to the actual name the attribute
            # with the connection id
            conn_id = getattr(self, self.conn_name_attr)
            conn_path = get_conn_path(conn_id)

            with log_sql_target_wrapper(self.get_uri(), conn_path, sql):
                # calling the original method to behave as the user expect
                return original_run(self, sql, *args, **kwargs)

        return original_run(self, sql, *args, **kwargs)

    return _run_sql



from collections import defaultdict
from itertools import chain
from typing import Dict, Iterable, List, Optional, Tuple

import attr

from more_itertools import first, padded

from dbnd._core.constants import DbndDatasetOperationType
from dbnd_snowflake.POC.sql_extract import Column, Schema
from targets.connections import build_conn_path
from targets.value_meta import ValueMeta
from targets.values import register_value_type
from targets.values.builtins_values import DataValueType


class Connection:
    @property
    def host(self) -> str:
        ...

    @property
    def port(self) -> Optional[int]:
        ...

    @property
    def database(self) -> Optional[str]:
        ...

    @property
    def schema(self) -> Optional[str]:
        ...


DTypes = Dict[str, str]


@attr.s
class SqlOperation:
    extracted_schema: Schema = attr.ib()
    dtypes: DTypes = attr.ib()
    records_count: int = attr.ib()
    query: str = attr.ib()
    query_id: str = attr.ib()
    success: bool = attr.ib()
    op_type: DbndDatasetOperationType = attr.ib()

    @property
    def columns(self) -> Optional[List[str]]:
        try:
            return list(self.dtypes)
        except TypeError:
            return None

    @property
    def columns_count(self) -> Optional[int]:
        try:
            return len(self.dtypes)
        except TypeError:
            return None

    @property
    def tables(self) -> Iterable[str]:
        return [col.table for col in self.extracted_columns]

    @property
    def extracted_columns(self) -> Iterable[Column]:
        return chain.from_iterable(self.extracted_schema.values())

    def evolve_schema(self, tables_schemas: Dict[str, DTypes]) -> "SqlOperation":
        if self.dtypes is not None:
            return self

        dtypes = self.build_dtypes(tables_schemas)
        return attr.evolve(self, dtypes=dtypes)

    def build_dtypes(self, tables_schemas: Dict[str, DTypes]) -> DTypes:
        dtypes = {}
        for col in self.extracted_columns:
            if col.is_wildcard:
                all_columns = tables_schemas[col.table]
                dtypes.update(all_columns)
            else:
                col_type = tables_schemas[col.table][col.name.lower()]
                dtypes[col.name] = col_type
        return dtypes

    def evolve_table_name(self, connection: Connection) -> "SqlOperation":
        schema: Schema = defaultdict(list)
        for name, cols in self.extracted_schema.items():
            for col in cols:
                table_name = render_table_name(connection, col.table)
                col = attr.evolve(col, table=table_name)
                schema[name].append(col)

        return attr.evolve(self, extracted_schema=schema)


def split_table_name(table_name: str) -> Tuple[str, Optional[str], Optional[str]]:
    table, schema, database = padded(reversed(table_name.split(".")[:3]), None, 3)
    return database, schema, table


def render_table_name(connection: Connection, table_name: str, sep: str = ".") -> str:
    database, schema, table = split_table_name(table_name)

    database = database or connection.database
    schema = schema or connection.schema

    path = sep.join(filter(None, [database, schema, table],))
    return path


def render_connection_path(
    connection: Connection, operation: SqlOperation, conn_type: str
) -> str:
    # We takes only the first table as a workaround
    table_name = first(operation.tables)
    path = render_table_name(connection, table_name, sep="/")
    return build_conn_path(
        conn_type=conn_type, hostname=connection.host, port=connection.port, path=path,
    )


class SqlOperationValueMeta(DataValueType):
    type = SqlOperation
    type_str = "SqlOperation"
    support_merge = False

    config_name = "sql_operation"
    is_lazy_evaluated = False

    def get_value_meta(self, value: SqlOperation, meta_conf):
        data_schema = {}
        data_dimensions = None

        if meta_conf.log_schema:
            data_schema = {
                "type": self.type_str,
                "dtypes": value.dtypes,
            }

        if meta_conf.log_size:
            data_dimensions = [value.records_count, value.columns_count]
            data_schema["shape"] = data_dimensions
            # todo: size?

        # currently stats and histogram are not supported
        stats, histograms = {}, {}
        hist_sys_metrics = None

        return ValueMeta(
            value_preview=None,
            data_dimensions=data_dimensions,
            data_schema=data_schema,
            data_hash=str(hash(self.to_signature(value))),
            descriptive_stats=stats,
            histogram_system_metrics=hist_sys_metrics,
            histograms=histograms,
        )


register_value_type(SqlOperationValueMeta())



from dbnd_snowflake.snowflake_resources import log_snowflake_resource_usage
from dbnd_snowflake.snowflake_table import log_snowflake_table
from dbnd_snowflake.snowflake_tracker import snowflake_query_tracker


__all__ = [
    "log_snowflake_resource_usage",
    "log_snowflake_table",
    "snowflake_query_tracker",
]



import contextlib
import functools

from collections import namedtuple
from itertools import chain, takewhile
from typing import Dict, List

import sqlparse

from snowflake.connector import SnowflakeConnection
from snowflake.connector.cursor import DictCursor, SnowflakeCursor

from dbnd import log_dataset_op
from dbnd_snowflake.POC.sql_extract import READ, WRITE, SqlQueryExtractor
from dbnd_snowflake.POC.sql_operation import (
    DTypes,
    SqlOperation,
    render_connection_path,
)


SNOWFLAKE_TYPE_CODE_MAP = {
    0: "NUMBER",
    1: "FLOAT",
    2: "VARCHAR",
    3: "DATE",
    4: "TIMESTAMP",
    5: "VARIANT",
    6: "TIMESTAMP_LTZ",
    7: "TIMESTAMP_TZ",
    8: "TIMESTAMP_TZ",
    9: "OBJECT",
    10: "ARRAY",
    11: "BINARY",
    12: "TIME",
    13: "BOOLEAN",
}
ParsedResultMetadata = namedtuple(
    "ParsedResultMetadata",
    [
        "name",
        "type_code",
        "display_size",
        "internal_size",
        "precision",
        "scale",
        "is_nullable",
    ],
)


def extract_schema_from_sf_desc(description):
    # todo: support ResultMetadata in version 2.4.6
    dtypes = {}
    for column in map(lambda col: ParsedResultMetadata(*col), description):
        dtypes[column.name] = SNOWFLAKE_TYPE_CODE_MAP[column.type_code]
    return dtypes


class PocSnowflakeTracker(object):
    def __init__(self):
        self.operations = []

    def __enter__(self):
        if not hasattr(SnowflakeCursor.execute, "__dbnd_patched__"):
            execute_original = SnowflakeCursor.execute

            @functools.wraps(execute_original)
            def snowflake_cursor_execute(cursor_self, *args, **kwargs):
                with self.track_execute(cursor_self, *args, **kwargs):
                    return execute_original(cursor_self, *args, **kwargs)

            snowflake_cursor_execute.__dbnd_patched__ = execute_original
            SnowflakeCursor.execute = snowflake_cursor_execute

        if not hasattr(SnowflakeConnection.close, "__dbnd_patched__"):
            close_original = SnowflakeConnection.close

            @functools.wraps(close_original)
            def snowflake_connection_close(connection_self, *args, **kwargs):
                # track connection before closing it (Example 1)
                self.unpatch_method(SnowflakeCursor, "execute")

                self.track_connection_data(connection_self)
                return close_original(connection_self, *args, **kwargs)

            snowflake_connection_close.__dbnd_patched__ = close_original
            SnowflakeConnection.close = snowflake_connection_close

        return self

    def unpatch_method(self, obj, original_attr, patched_attr="__dbnd_patched__"):
        method = getattr(obj, original_attr)
        if hasattr(method, patched_attr):
            setattr(
                SnowflakeCursor, original_attr, getattr(method, patched_attr),
            )

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unpatch_method(SnowflakeCursor, "execute")
        self.unpatch_method(SnowflakeConnection, "close")

    def track_connection_data(self, connection_self: SnowflakeConnection):
        # update the tables names
        operations = [op.evolve_table_name(connection_self) for op in self.operations]

        # looks for tables schemas
        tables = chain.from_iterable(op.tables for op in operations)
        tables_schemas: Dict[str, DTypes] = {
            table: get_snowflake_table_schema(connection_self, table)
            for table in tables
        }

        operations: List[SqlOperation] = [
            op.evolve_schema(tables_schemas) for op in operations
        ]

        for op in operations:
            log_dataset_op(
                op_path=render_connection_path(connection_self, op, "snowflake"),
                op_type=op.op_type,
                success=op.success,
                data=op,
                with_schema=True,
                send_metrics=True,
            )

    @contextlib.contextmanager
    def track_execute(self, cursor, command, *args, **kwargs):
        success = True
        try:
            yield
        except Exception as e:
            success = False
            raise
        finally:
            operations = build_snowflake_operations(cursor, command, success)
            self.operations.extend(operations)


def get_snowflake_table_schema(connection, table):
    desc_results = snowflake_query(connection, f"desc table {table}")
    schema = {}
    for col_desc in desc_results:
        # remove the trailing part of the column type that's inside brackets
        # `NUMBER(123,0) -> NUMBER`, `VARCHAR(23,2) -> VARCHAR` and so on
        normalized_col_type = "".join(takewhile(lambda c: c != "(", col_desc["type"]))
        normalized_col_name = col_desc["name"].lower()
        schema[normalized_col_name] = normalized_col_type
    return schema


def snowflake_query(connection: SnowflakeConnection, query: str, params=None):
    try:
        with connection.cursor(DictCursor) as cursor:
            cursor.execute(query, params)
            result = cursor.fetchall()
            return result
    except Exception:
        print("Error occurred during querying Snowflake, query: %s", query)
        raise


def build_snowflake_operations(
    cursor: SnowflakeCursor, command: str, success: bool
) -> List[SqlOperation]:
    operations = []

    # find the relevant operations schemas from the command
    parsed_query = sqlparse.parse(command)[0]
    extracted = SqlQueryExtractor().extract_operations_schemas(parsed_query)

    # helper method for building an operation from common values
    build_operation = functools.partial(
        SqlOperation,
        records_count=cursor.rowcount,
        query=command,
        query_id=cursor.sfqid,
        success=success,
    )

    if READ in extracted and WRITE not in extracted:
        # In this case this is a read only operation which means the cursor holds the actual result and the
        # description contains the schema of the operation.
        schema = extract_schema_from_sf_desc(cursor.description)
        read = build_operation(
            extracted_schema=extracted[READ], dtypes=schema, op_type=READ
        )
        operations.append(read)
    else:
        # This is write operation which means the cursor holds only the `effected_rows` result which holds no
        # schema.
        if READ in extracted:
            read = build_operation(
                extracted_schema=extracted[READ], dtypes=None, op_type=READ
            )
            operations.append(read)

        write = build_operation(
            extracted_schema=extracted[WRITE], dtypes=None, op_type=WRITE
        )
        operations.append(write)
    return operations



from collections import defaultdict
from typing import Dict, List, Optional, Tuple, Union

import attr

from sqlparse.sql import (
    Comparison,
    Function,
    Identifier,
    IdentifierList,
    Parenthesis,
    Token,
    TokenList,
)
from sqlparse.tokens import Keyword, Name, Token as TokenType, Whitespace, Wildcard

from dbnd._core.constants import DbndTargetOperationType
from dbnd_snowflake.POC.utils import ddict2dict


@attr.s
class Column:
    table = attr.ib()
    name = attr.ib(default=None)
    alias = attr.ib(default=None)

    @property
    def is_wildcard(self):
        return self.name == "*"


Columns = List[Column]
Schema = Dict[str, Columns]

OP_TYPE = DbndTargetOperationType
READ = DbndTargetOperationType.read
WRITE = DbndTargetOperationType.write
READ_OPERATIONS = {
    "FROM",
    "LEFT JOIN",
    "JOIN",
    "INNER JOIN",
    "LEFT OUTER JOIN",
    "RIGHT JOIN",
    "RIGHT OUTER",
    "RIGHT OUTER JOIN",
}


class SqlQueryExtractor:
    def extract_operations_schemas(self, statement: TokenList) -> Dict[OP_TYPE, Schema]:
        """
        This function go over the statement and extract the used tables names and the operation that it is been done
        with them. The result is a list of pairs where the left element is the table name and the right one is the
        operation.
        """
        # `sqlparse` implementation details:
        # Using the changing `idx` (the index of the current token) to keep moving in the token list (statement),
        # after finding a current token. First idx is -1 because this is the first position outside the token list.
        # Each call to `token_next` or `token_next_by` using the current idx to start searching from the current
        # position
        idx = -1

        # holds the columns we find in the scope and not yet assigned to any schema
        columns: Columns = []
        # holds any dynamic table schema that is created in the scope.
        # dynamic tables are created by CTE statements
        dynamic_tables: Dict[str, Schema] = {}

        # collect the operations and their schemas
        operations = defaultdict(lambda: defaultdict(list))
        while True:
            # each iteration we pick up the next token we didn't handle yet
            idx, token = self._next_non_empty_token(idx, statement)
            if not token:
                # done iterating over the sql statement
                # converting the result to regular dict for convince
                return ddict2dict(operations)

            token: Token
            if token.ttype in Keyword:
                operation_name = token.value.upper()
                idx_potential, next_token = self._next_non_empty_token(idx, statement)

                if operation_name == "SELECT":
                    extracted_columns, idx = self.handle_select(
                        idx_potential, next_token
                    )
                    columns.extend(extracted_columns)

                elif operation_name == "INTO":
                    extracted, idx = self.handle_into(
                        idx_potential, next_token, statement
                    )
                    operations[WRITE].update(extracted)

                elif operation_name == "UPDATE":
                    extracted, idx = self.handle_update(
                        idx_potential, next_token, statement
                    )
                    operations[WRITE].update(extracted)

                elif operation_name in READ_OPERATIONS:
                    extracted, idx, columns = self.handle_read_operation(
                        idx_potential, next_token, columns, dynamic_tables
                    )
                    # to support UNION we extend the array of the same column
                    for k, v in extracted.items():
                        operations[READ][k].extend(v)

                elif token.ttype is Keyword.CTE:
                    # after a CTE statement (WITH) we expect identifiers list with each has a sub-query inside parenthesis
                    idx, token = self._next_non_empty_token(idx, statement)
                    for identifier in token.get_identifiers():
                        _, cte_statement = identifier.token_next_by(i=Parenthesis)
                        cte_operations = self.extract_operations_schemas(cte_statement)
                        extracted = self.enrich_with_dynamic(
                            cte_operations[READ], dynamic_tables
                        )
                        dynamic_tables[identifier.get_name()] = extracted

    def handle_select(self, idx: int, next_token: Token) -> Tuple[Columns, int]:
        tokens: List[Token] = (
            next_token.get_identifiers()
            if isinstance(next_token, IdentifierList)
            else [next_token]
        )
        columns = self._extract_columns(tokens)
        return columns, idx

    def handle_read_operation(
        self,
        idx: int,
        next_token: Token,
        columns: Columns,
        dynamic_tables: Dict[str, Schema],
    ) -> Tuple[Schema, int, Columns]:
        # looking for a sub-query first - it can be the current token or the first son of this token
        extracted = {}
        if isinstance(next_token, Parenthesis):
            # sub query doesnt have an alias name:
            # `... from (<sub_query>) where ...`
            extracted = self.extract_from_subquery(next_token, "anon", columns)

        elif isinstance(next_token, TokenList) and isinstance(
            next_token.token_first(), Parenthesis
        ):
            # the sub query has an alias name:
            # `... from (<sub_query>) as <alias_name> where ...`
            extracted = self.extract_from_subquery(
                next_token.token_first(), next_token.get_alias(), columns
            )

        elif next_token.ttype not in Keyword:
            # no subquery - just parse the source/dest name of the operator
            table_alias = self.get_identifier_name(next_token)
            table_name = self.get_full_name(next_token)
            extracted, left_columns = self.extract_schema(
                table_name, table_alias, columns
            )
            columns = left_columns
            extracted = self.enrich_with_dynamic(extracted, dynamic_tables)

        return extracted, idx, columns

    def handle_into(
        self, idx: int, next_token: Token, statement: TokenList
    ) -> Tuple[Schema, int]:
        table_alias = self.get_identifier_name(next_token)
        table_name = self.get_full_name(next_token)
        if isinstance(next_token, Function):
            extracted = self.extract_write_schema(
                next_token.get_parameters(), table_alias, table_alias
            )
        else:
            extracted = self.extract_write_schema(
                [Token(Wildcard, "*")], table_name, table_alias
            )
        return extracted, idx

    def handle_update(
        self, idx: int, next_token: Token, statement: TokenList
    ) -> Tuple[Schema, int]:
        table_alias = next_token.get_name()
        table_name = next_token.normalized

        idx, token = self._next_non_empty_token(idx, statement)
        operation_name = token.value.upper()

        extracted = {}
        if operation_name == "SET":
            idx, token = self._next_non_empty_token(idx, statement)
            extracted = self.extract_write_schema(
                token.get_identifiers(), table_name, table_alias
            )

        return extracted, idx

    def extract_write_schema(
        self, tokens: List[Token], table_name: str, table_alias: str
    ) -> Schema:
        cols = self._extract_columns(tokens)
        extracted, _ = self.extract_schema(table_name, table_alias, cols)
        return extracted

    def extract_from_subquery(
        self, tok: TokenList, table_alias: str, columns: Columns
    ) -> Schema:
        dynamic = self.extract_operations_schemas(tok)
        extracted, left_columns = self.extract_schema(table_alias, table_alias, columns)
        extracted = self.enrich_with_dynamic(extracted, {table_alias: dynamic[READ]})
        return extracted

    def extract_schema(
        self, table_name: str, table_alias: str, columns: Columns
    ) -> Tuple[Schema, Columns]:
        collected_schema = defaultdict(list)
        left_columns = []

        col: Column
        for col in columns:
            if col.is_wildcard:
                ev_col = attr.evolve(col, table=table_name, alias=f"{table_name}.*")
                collected_schema[ev_col.alias].append(ev_col)
                left_columns.append(col)

            elif col.table and col.table != table_alias:
                left_columns.append(col)

            else:
                ev_col = attr.evolve(col, table=table_name)
                collected_schema[ev_col.alias].append(ev_col)

        return collected_schema, left_columns

    @staticmethod
    def _next_non_empty_token(
        idx: int, token_list: TokenList
    ) -> Tuple[Optional[int], Optional[Token]]:
        return token_list.token_next(idx, skip_ws=True, skip_cm=True)

    def enrich_with_dynamic(
        self, extracted: Schema, dynamic: Dict[str, Schema],
    ) -> Schema:
        if not dynamic:
            return extracted

        result = defaultdict(list)
        for name, cols in extracted.items():
            for col in cols:
                if col.table in dynamic:
                    dynamic_table_schema = dynamic[col.table]
                    if col.is_wildcard:
                        result.update(dynamic_table_schema)

                    elif col.name in dynamic_table_schema:
                        result[col.name].extend(dynamic_table_schema[col.name])

                    else:
                        result[col.name].append(col)
                else:
                    result[name].append(col)
        return result

    def _extract_columns(self, tokens: List[Token]) -> Columns:
        collected_columns = []
        for token in tokens:
            if token.ttype == Wildcard:
                col = Column(table=None, name="*", alias=None)

            elif isinstance(token, Identifier):
                name = token.get_real_name()
                alias = token.get_alias() or name

                first = token.token_first()
                ref = (
                    first.value
                    if len(token.tokens) > 1 and first.ttype == Name
                    else None
                )
                col = Column(table=ref, name=name, alias=alias)

            elif isinstance(token, Comparison):
                name = token.left.get_real_name()
                col = Column(table=None, name=name, alias=name)

            else:
                name = token.get_alias() or token.normalized
                col = Column(table=None, name=name, alias=name)

            collected_columns.append(col)
        return collected_columns

    @staticmethod
    def get_identifier_name(identifier):
        return identifier.get_alias() or identifier.value.split(".")[-1]

    @staticmethod
    def get_full_name(identifier):
        # "name as alias" or "name alias" or "complicated column expression alias"
        ws_index, ws = identifier.token_next_by(t=Whitespace)
        if len(identifier.tokens) > 2 and ws is not None:
            return "".join(token.value for token in identifier.tokens[:ws_index])

        return identifier.value
